import os


def expand_env(value):
    """
    Expand environment and user variables

    :param str value: Value to expand
    :return: Expanded value
    :rtype: str
    """
    return os.path.expandvars(os.path.expanduser(value))



import json
import os
import six

from .exceptions import InvalidConfigException, MissingEnvironmentVariableException


def set_config_defaults(config):
    """
    Set default values

    :param dict config: Configuration
    :return:
    """
    config.setdefault('OutputDirectory', '$HOME/gossamer')
    config.setdefault('OutputFile', 'goss-generated-aliases.sh')
    config.setdefault('GossamerPath', '/usr/local/bin/gossamer')
    config.setdefault('AWSCLIPath', '/usr/local/bin/aws')
    config.setdefault('AWSCredentialsPath', '$HOME/.aws/credentials')
    config.setdefault('BaseProfile', 'default')


def validate_config(config):
    """
    Validate the config file

    :param dict config: Config
    :raises InvalidConfigException
    """
    required_fields = {
        'Accounts': list
    }

    if six.PY2:
        str_types = (unicode, str)
    else:
        str_types = str

    optional_fields = {
        'OutputDirectory': str_types,
        'OutputFile': str_types,
        'GossamerPath': str_types,
        'AWSCredentialsPath': str_types,
        'BaseProfile': str_types,
        'RoleAliases': dict
    }

    # Validate required fields
    for field, field_type in required_fields.items():
        if field not in config:
            raise InvalidConfigException('Missing key "%s" in configuration file.' % field)
        elif not isinstance(config[field], field_type):
            raise InvalidConfigException('Key "%s" should be type %s' % (field, field_type))

    # Validate optional fields
    for field, field_type in optional_fields.items():
        if field in config and not isinstance(config[field], field_type):
            raise InvalidConfigException('Key "%s" should be type %s' % (field, field_type))

    # Make sure user has $MFA in their profile
    if 'MFA' not in os.environ:
        raise MissingEnvironmentVariableException(
            'The environment variable $MFA is missing. '
            'This should be configured with your multi-factor device serial number.'
        )


def get_config(config_file):
    """
    Read and validate configuration file

    :param str config_file: Config file path
    :return: Configuration
    :rtype: dict
    """
    # Check if file exists
    if not os.path.exists(config_file):
        raise InvalidConfigException('The configuration file "%s" does not exist' % config_file)

    # Read config file
    with open(config_file, 'r') as f:
        config = json.loads(f.read())

    # Configure default values
    set_config_defaults(config)

    # Validate configuration
    validate_config(config)

    return config



import json
import re


def import_role_files(role_files):
    """
    Import Gossamer role files to generate a goss-gen config file

    :param list role_files: List of role file paths
    :return: Generated configuration
    :rtype: dict
    """

    account_roles = set()
    accounts = []

    for filename in role_files:
        with open(filename, 'r') as f:
            file_data = json.loads(f.read())

        for item in file_data['Roles']:
            matches = re.match(r'arn:aws:iam::(?P<account>\d+):role/(?P<role>.+)', item['RoleArn'])

            account = matches.group('account')
            role = matches.group('role')

            if (account, role) in account_roles:
                continue

            accounts.append({
                'Id': account,
                'Name': item['AccountName'],
                'Region': item['Region'],
                'Role': role
            })

            account_roles.add((account, role))

    return {'Accounts': accounts}



import json
import os
import stat
from argparse import ArgumentParser

from .config import get_config
from .role_import import import_role_files
from .utils import expand_env


def main():
    parser = ArgumentParser(
        description='Generate Gossamer config files and aliases'
    )

    parser.add_argument(
        '--config-file',
        '-c',
        help='Configuration file to use. Defaults to $HOME/goss-config.json [$GOSS_GEN_CONFIG]',
        metavar='<path>'
    )
    
    parser.add_argument(
        '--import-file',
        '-i',
        help='Import an existing gossamer role file and output a configuration file. '
             'This argument can be specified multiple times for multiple gossamer role files.',
        action='append'
    )

    options = parser.parse_args()

    # Configuration file
    if options.config_file:
        config_file = options.config_file
    else:
        config_file = expand_env(os.environ.get('GOSS_GEN_CONFIG', os.path.join('$HOME', 'goss-config.json')))

    # Expand absolute path
    config_file = os.path.abspath(config_file)

    # Check for role file imports
    if options.import_file:
        config = import_role_files(options.import_file)

        with open(config_file, 'w') as f:
            f.write(json.dumps(config, indent=4))

        print('Configuration file written to "%s"' % config_file)
        return

    # Read in the config
    config = get_config(config_file)

    # Build output file path
    output_dir = expand_env(config['OutputDirectory'])
    output_path = os.path.join(
        output_dir,
        config['OutputFile']
    )

    # Create directory if needed
    try:
        os.makedirs(output_dir)
    except OSError:
        pass

    # Read in data
    roles = {}
    aliases = []
    account_data = config['Accounts']
    role_aliases = config.get('RoleAliases', {})

    # Loop through all accounts configured
    for item in account_data:
        # Initialize role dict
        roles.setdefault(item['Role'], [])

        # Create alias
        alias = item.get('Alias', '')
        if not alias:
            alias = item['Name']

        # Add role to the dict
        roles[item['Role']].append({
            "RoleArn": "arn:aws:iam::%(account_id)s:role/%(role)s" % {
                'account_id': item['Id'],
                'role': item['Role']
            },
            "AccountName": item['Name'],
            "Region": item['Region']
        })

        # Write AWS alias
        aliases.append(
            "alias aws-%(alias)s='%(aws_cli_path)s --profile %(account_name)s'\n" % {
                'alias': alias,
                'aws_cli_path': config['AWSCLIPath'],
                'account_name': item['Name']
            }
        )

        # Write AWS insecure alias
        aliases.append(
            "alias awsi-%(alias)s='%(aws_cli_path)s --profile %(account_name)s --no-verify-ssl'\n" % {
                'alias': alias,
                'aws_cli_path': config['AWSCLIPath'],
                'account_name': item['Name']
            }
        )

        # Write gossamer alias
        aliases.append(
            "alias goss-%(alias)s='%(gossamer_path)s -a arn:aws:iam::%(account_id)s:role/%(role)s -profile %(profile)s "
            "-serialnumber $MFA -o %(aws_creds_path)s -entryname %(account_name)s -force -tokencode'\n" % {
                'alias': alias,
                'gossamer_path': config['GossamerPath'],
                'role': item['Role'],
                'profile': config['BaseProfile'],
                'aws_creds_path': config['AWSCredentialsPath'],
                'account_id': item['Id'],
                'account_name': item['Name']
            }
        )

    # Write role files
    for role_name, role_data in roles.items():
        normalized_role_name = role_name.replace('/', '-')

        role_file = os.path.join(
            output_dir,
            normalized_role_name + '.json'
        )
        output_data = {'Roles': role_data}

        with open(role_file, 'w') as f:
            f.write(json.dumps(output_data, indent=4))

        alias = role_aliases.get(role_name, normalized_role_name)

        # Write gossamer alias
        aliases.append(
            "alias goss-%(alias)s='%(gossamer_path)s -rolesfile %(role_file)s -profile %(profile)s -serialnumber $MFA "
            "-o %(aws_creds_path)s -force -tokencode'\n" % {
                'alias': alias,
                'gossamer_path': config['GossamerPath'],
                'role_file': role_file,
                'profile': config['BaseProfile'],
                'aws_creds_path': config['AWSCredentialsPath']
            }
        )

    # Write alias file
    with open(output_path, 'w') as f:
        f.writelines(aliases)

    # Make the file executable
    os.chmod(output_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IROTH)

    print('Generated Gossamer aliases successfully.')



class InvalidConfigException(Exception):
    """Raised when the configuration is not valid."""


class MissingEnvironmentVariableException(Exception):
    """Raised when an environment variable is missing."""



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='goss-config-gen',
    description='Generate Gossamer config files and AWS aliases',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='gossamer config aws',
    url='https://github.com/GESkunkworks/goss-config-gen',
    use_scm_version=True,
    author='Michael Palmer',
    author_email='github@michaeldpalmer.com',
    packages=find_packages(),
    setup_requires=[],
    install_requires=[
        'setuptools_scm~=3.1.0',
        'six~=1.11.0'
    ],
    entry_points={
        'console_scripts': [
            'goss-config-gen = goss_config_gen.cli:main',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 5 - Production/Stable',
        'Topic :: Utilities'
    ]
)
